// Klasse, um an eine Rckrufadresse eine benutzerspezifische Adresse
// zu liefern; Anwendungsbeispiel siehe unten - h#s 12/05

#ifndef THUNK_H
#define THUNK_H

#include <windows.h>		// Datentypen (sonst nicht notwendig)

#ifdef _M_IX86
# define THUNK_CODESIZE 12	// Code-Gre ist plattformabhngig!
#else
# error Nicht untersttzte Plattform! Ggf. Kode hier einfgen!
#endif

class CThunk {
public:
 MakeCdecl  (LPVOID FuncPtr, PVOID Arg); // fr cdecl-Funktionen
 MakeStdcall(LPVOID FuncPtr, PVOID Arg); // fr stdcall-Funktionen
private:
 BYTE code[THUNK_CODESIZE];
};

/* Zweck:
 * Es kommt immer wieder vor, dass es Rckruffunktionen ohne
 * frei verwendbaren Hilfsparameter gibt.
 * Dann ist es bei Verwendung mehrerer Daten-Instanzen nicht mehr
 * mglich, zwischen "Objekt A" und "Objekt B" zu unterscheiden,
 * auer ber eine globale Variable. Dann ist's nicht mehr multi-
 * thread-sicher. Da hilft blo noch eine thread-lokale Variable.
 *
 * Sofern jedoch wenigstens die Rckrufadresse frei whlbar ist,
 * kann diese auf dynamisch erzeugten Kode zeigen, welche den
 * Zusatzparameter nachreicht. Ein Thunk eben.
 * Geeignet fr alle von-Neumann-Prozessorarchitekturen.
 */

/****************************************************************
 * Verwendung: Beispiel qsort:					*
 *								*
 * // Nicht-Memberfunktion (Freund von CBeispiel):		*
 * static int _cdecl CompFunc(					*
 *  CBeispiel* obj, LPVOID, void*left, void*right) {		*
 *  return obj->Vergleich(left,right)				*
 * }								*
 *								*
 * // Beispiel-Klasse						*
 * class CBeispiel {						*
 *  friend int _cdecl CompFunc(CExample*,LPVOID,void*,void*);	*
 *  CThunk m_cThunk;						*
 *  CBeispiel() {m_cThunk.MakeCdecl(CompFunc, this);}		*
 *  int Vergleich(void*left, void*right) {return ...;}		*
 *  void Sortiere() {qsort(...Array, ...Size, &m_cThunk);}	*
 * }								*
 ****************************************************************/
 
/* Verwendung (allgemein):
 * Mit MakeCdecl() wird das CThunk-Objekt mit Kode gefllt, welches
 * eine Funktion mit der gleichen Signatur aufruft wie ohne Thunk,
 * mit zwei zustzlichen Parametern am Anfang:
 * -> 1. Parameter = bergebener Parameter
 * -> 2. Parameter = (fest) LPVOID; dieser darf nicht angerhrt werden!
 * Die Funktion muss mit _cdecl deklariert werden!
 *
 * Mit MakeStdcall() wird das CThunk-Objekt mit Kode gefllt, welches
 * eine Funktion mit der gleichen Signatur aufruft wie ohne Thunk,
 * mit einem zustzlichen Parameter am Anfang:
 * -> 1. Parameter = bergebener Parameter
 * Die Funktion muss mit _stdcall deklariert werden!
 *
 * Mit MakeStdcall() knnen auch _pascal-Funktionen angesprungen werden,
 * sofern vom Compiler untersttzt. Der Extra-Parameter erscheint
 * als letzter Parameter.
 *
 * Es gibt hier kein MakeFastcall() - eine solche Aufrufkonvention ist
 * nicht allgemein festgelegt.
 *
 * Weil bei (nicht-statischen) Memberfunktionen der this-Zeiger als erstes
 * bergeben wird, sollte auch der direkte Aufruf von Objektfunktionen
 * mglich sein; dies ist ungetestet.
 * Der Aufruf einer virtuellen Funktion ist dabei nicht direkt mglich.
 */

#endif
